@@ -189,7 +189,8 @@ fn build_runtime_function(
           }
           rt::CTR => {
             //println!("Didn't match because of CTR. i={} {} {}", i, rt::get_tag(rt::ask_arg(mem, term, i)), rt::get_val(*cond));
-            let same_tag = rt::get_tag(rt::ask_arg(mem, term, i)) == rt::CTR;
+            let ctor_tag = rt::get_tag(rt::ask_arg(mem, term, i));
+            let same_tag = ctor_tag == rt::CTR || ctor_tag == rt::CAL;
             let same_ext = rt::get_ext(rt::ask_arg(mem, term, i)) == rt::get_ext(*cond);
             matched = matched && same_tag && same_ext;
           }
@@ -523,7 +524,7 @@ fn alloc_term(
 }
 
 // Evaluates a Lambolt term to normal form
-pub fn eval_code(call: &lang::Term, code: &str, debug: bool) -> Result<(Box<lang::Term>, u64, u64, u64), String> {
+pub fn eval_code(call: &lang::Term, code: &str, debug: bool) -> Result<(String, u64, u64, u64), String> {
   // Creates a new Runtime worker
   let mut worker = rt::new_worker();
 
@@ -545,8 +546,12 @@ pub fn eval_code(call: &lang::Term, code: &str, debug: bool) -> Result<(Box<lang
   let time = init.elapsed().as_millis() as u64;
 
   // Reads it back to a Lambolt string
-  let norm = rd::as_term(&worker, &Some(book), host)?;
+  let book = Some(book);
+  let code = match rd::as_term(&worker, &book, host) {
+    Ok(x)   => format!("{}", x),
+    Err(..) => rd::as_code(&worker, &book, host),
+  };
 
   // Returns the normal form and the gas cost
-  Ok((norm, worker.cost, worker.size, time))
+  Ok((code, worker.cost, worker.size, time))
 }
