@@ -264,7 +264,7 @@ pub fn parse_ctr(state: parser::State) -> parser::Answer<Option<BTerm>> {
     Box::new(|state| {
       let (state, _) = parser::text("(", state)?;
       let (state, head) = parser::get_char(state)?;
-      Ok((state, ('A'..='Z').contains(&head) || head == '.'))
+      Ok((state, ('A'..='Z').contains(&head)))
     }),
     Box::new(|state| {
       let (state, open) = parser::text("(", state)?;
@@ -354,7 +354,7 @@ pub fn parse_var(state: parser::State) -> parser::Answer<Option<BTerm>> {
   parser::guard(
     Box::new(|state| {
       let (state, head) = parser::get_char(state)?;
-      Ok((state, ('a'..='z').contains(&head) || head == '_'))
+      Ok((state, ('a'..='z').contains(&head) || head == '_' || head == '$'))
     }),
     Box::new(|state| {
       let (state, name) = parser::name(state)?;
