@@ -318,7 +318,6 @@ pub fn reduce(
     if debug {
       println!("------------------------");
       println!("{}", show_term(mem, ask_lnk(mem, 0), _opt_id_to_name, term));
-      //println!("{}", show_mem(mem));
     }
 
     if init == 1 {
@@ -433,8 +432,7 @@ pub fn reduce(
               inc_cost(mem);
               subst(mem, ask_arg(mem, term, 0), ask_arg(mem, arg0, 0));
               subst(mem, ask_arg(mem, term, 1), ask_arg(mem, arg0, 1));
-              let _done =
-                link(mem, host, ask_arg(mem, arg0, if get_tag(term) == DP0 { 0 } else { 1 }));
+              let _done = link(mem, host, ask_arg(mem, arg0, if get_tag(term) == DP0 { 0 } else { 1 }));
               clear(mem, get_loc(term, 0), 3);
               clear(mem, get_loc(arg0, 0), 2);
               init = 1;
@@ -495,6 +493,14 @@ pub fn reduce(
               let done = Ctr(arit, func, if get_tag(term) == DP0 { ctr0 } else { ctr1 });
               link(mem, host, done);
             }
+          } else if get_tag(arg0) == ERA {
+            inc_cost(mem);
+            subst(mem, ask_arg(mem, term, 0), Era());
+            subst(mem, ask_arg(mem, term, 1), Era());
+            link(mem, host, Era());
+            clear(mem, get_loc(term, 0), 3);
+            init = 1;
+            continue;
           }
         }
         OP2 => {
@@ -566,7 +572,6 @@ pub fn reduce(
           let _ari = get_ari(term);
           if let Some(f) = &funcs[fun as usize] {
             if (f.rewriter)(mem, dups, host, term) {
-              //println!("cal-fun");
               init = 1;
               continue;
             }
@@ -653,9 +658,19 @@ pub fn normal(
   opt_id_to_name: Option<&HashMap<u64, String>>,
   debug: bool,
 ) -> Lnk {
-  let mut seen = vec![0; 4194304];
+  let mut done;
   let mut dups = 0;
-  normal_go(mem, &mut dups, funcs, host, &mut seen, opt_id_to_name, debug)
+  let mut cost = mem.cost;
+  loop {
+    let mut seen = vec![0; 4194304];
+    done = normal_go(mem, &mut dups, funcs, host, &mut seen, opt_id_to_name, debug);
+    if mem.cost != cost {
+      cost = mem.cost;
+    } else {
+      break;
+    }
+  }
+  done
 }
 
 // Debug
@@ -689,7 +704,7 @@ pub fn show_lnk(x: Lnk) -> String {
       F32 => "F32",
       OUT => "OUT",
       NIL => "NIL",
-      _ => "???",
+      _   => "?",
     };
     format!("{}{}:{:x}:{:x}", tgs, ari, ext, val)
   }
@@ -778,16 +793,16 @@ pub fn show_term(
   ) -> String {
     let done = match get_tag(term) {
       DP0 => {
-        format!("a{}", names.get(&get_loc(term, 0)).unwrap_or(&String::from("?")))
+        format!("a{}", names.get(&get_loc(term, 0)).unwrap_or(&String::from("?a")))
       }
       DP1 => {
-        format!("b{}", names.get(&get_loc(term, 0)).unwrap_or(&String::from("?")))
+        format!("b{}", names.get(&get_loc(term, 0)).unwrap_or(&String::from("?b")))
       }
       VAR => {
-        format!("x{}", names.get(&get_loc(term, 0)).unwrap_or(&String::from("?")))
+        format!("x{}", names.get(&get_loc(term, 0)).unwrap_or(&String::from("?c")))
       }
       LAM => {
-        let name = format!("x{}", names.get(&get_loc(term, 0)).unwrap_or(&String::from("?")));
+        let name = format!("x{}", names.get(&get_loc(term, 0)).unwrap_or(&String::from("?d")));
         format!("λ{} {}", name, go(mem, ask_arg(mem, term, 1), names, opt_id_to_name, focus))
       }
       APP => {
@@ -806,23 +821,23 @@ pub fn show_term(
         let val0 = go(mem, ask_arg(mem, term, 0), names, opt_id_to_name, focus);
         let val1 = go(mem, ask_arg(mem, term, 1), names, opt_id_to_name, focus);
         let symb = match oper {
-          0x00 => "+",
-          0x01 => "-",
-          0x02 => "*",
-          0x03 => "/",
-          0x04 => "%",
-          0x05 => "&",
-          0x06 => "|",
-          0x07 => "^",
-          0x08 => "<<",
-          0x09 => ">>",
-          0x10 => "<",
-          0x11 => "<=",
-          0x12 => "=",
-          0x13 => ">=",
-          0x14 => ">",
-          0x15 => "!=",
-          _ => "?",
+          0x0 => "+",
+          0x1 => "-",
+          0x2 => "*",
+          0x3 => "/",
+          0x4 => "%",
+          0x5 => "&",
+          0x6 => "|",
+          0x7 => "^",
+          0x8 => "<<",
+          0x9 => ">>",
+          0xA => "<",
+          0xB => "<=",
+          0xC => "=",
+          0xD => ">=",
+          0xE => ">",
+          0xF => "!=",
+          _   => "?e",
         };
         format!("({} {} {})", symb, val0, val1)
       }
@@ -835,7 +850,7 @@ pub fn show_term(
         let args: Vec<String> =
           (0..arit).map(|i| go(mem, ask_arg(mem, term, i), names, opt_id_to_name, focus)).collect();
         let name = if let Some(id_to_name) = opt_id_to_name {
-          id_to_name.get(&func).unwrap_or(&String::from("?")).clone()
+          id_to_name.get(&func).unwrap_or(&String::from("?f")).clone()
         } else {
           format!(
             "{}{}",
@@ -845,7 +860,10 @@ pub fn show_term(
         };
         format!("({}{})", name, args.iter().map(|x| format!(" {}", x)).collect::<String>())
       }
-      _ => String::from("?"),
+      ERA => {
+        format!("*")
+      }
+      _ => format!("?g({})", get_tag(term)),
     };
     if term == focus {
       format!("${}", done)
@@ -857,13 +875,11 @@ pub fn show_term(
   let mut text = go(mem, term, &names, opt_id_to_name, focus);
   for (_key, pos) in lets {
     // todo: reverse
-    let what = String::from("?");
+    let what = String::from("?h");
     //let kind = kinds.get(&key).unwrap_or(&0);
     let name = names.get(&pos).unwrap_or(&what);
-    let nam0 =
-      if ask_lnk(mem, pos + 0) == Era() { String::from("*") } else { format!("a{}", name) };
-    let nam1 =
-      if ask_lnk(mem, pos + 1) == Era() { String::from("*") } else { format!("b{}", name) };
+    let nam0 = if ask_lnk(mem, pos + 0) == Era() { String::from("*") } else { format!("a{}", name) };
+    let nam1 = if ask_lnk(mem, pos + 1) == Era() { String::from("*") } else { format!("b{}", name) };
     text.push_str(&format!(
       "\ndup {} {} = {};",
       //kind,
